import pytest

from PyEtalon.etalon import Etalon
import numpy as np
from PyEtalon.custom_material_data import Vacuum, Suprasil, Silver
from PyEtalon import plotting


@pytest.fixture(scope="module")
def silver_etalon():
    return Etalon(
        [Vacuum, Silver, Suprasil],  # specify the layer materials
        ["Vacuum", "Silver", "Substrate"],  # specify the layer names
        [-np.inf, 30.0, np.inf],  # specify the layer thicknesses
        [0, 1, 2],  # specify the layer order/indices
        wavelength=np.linspace(500, 1100, 5000),  # specify the wavelength range
        d_spacer=5e-3,  # specify the spacer thickness in meters
        aoi=0.0,  # specify the angle of incidence in degrees
        identifier="Silver 30nm",  # specify the identifier of the etalon
    )


def test_etalon():
    pass


def test_reflectivity_transmissivity():
    e = Etalon(
        [Vacuum, Suprasil],
        ["vacuum", "Suprasil"],
        [-np.inf, np.inf],
        [0, 1],
        wavelength=np.array([587.56]),
        d_spacer=0.01,
        aoi=20.0,
        identifier="test",
    )
    ref = e.mirror_reflectivity
    trans = e.mirror_transmission
    # data checked on https://www.rp-photonics.com/fresnel_equations.html
    assert np.isclose(ref, 0.0411, atol=1e-4)
    assert np.isclose(trans, 0.959, atol=1e-3)


def test_etalon_fsr(silver_etalon):
    assert silver_etalon.ideal_fsr == 29.9792458


def test_plotting(silver_etalon):
    plotting.plot_gd(silver_etalon)
    plotting.plot_fsr(silver_etalon)
    plotting.plot_gdd(silver_etalon)
    plotting.plot_phase(silver_etalon)
    plotting.plot_absorption(silver_etalon)
    plotting.plot_layer_thickness_influence(silver_etalon)
    plotting.plot_layer_thickness_influence_gd(silver_etalon)
    plotting.plot_layer_thickness_influence_reflectivity(silver_etalon)
    plotting.plot_normalized_fsr(silver_etalon)
    plotting.plot_stack([silver_etalon])
    plotting.plot_absorption(silver_etalon)
    plotting.plot_normalized_fsr(silver_etalon)
    plotting.plot_mirror_transmission(silver_etalon)
    plotting.plot_reflectivity(silver_etalon)
    plotting.plot_reflectivity_finesse(silver_etalon)
    plotting.plot_refractive_index_materials(
        silver_etalon, silver_etalon.wavelength / 1000.0
    )
    plotting.plot_refractive_index_materials(
        silver_etalon, silver_etalon.wavelength / 1000.0, False
    )



from typing import List, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches, cm

from PyEtalon.etalon import Etalon

c = 299792458  # speed of light [m/s]


def plot_phase(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("Unwrapped Phase Shift of coating")
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("phase shift [rad]")
    for ee in e:
        plt.plot(ee.wavelength, ee.phase, label=ee.identifier)
    plt.legend()


def plot_gd(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("GD of coating")
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("GD")
    for ee in e:
        plt.plot(ee.wavelength, ee.gd, label=ee.identifier)
    plt.legend()


def plot_gdd(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("GDD of coating")
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("GDD")
    for ee in e:
        plt.plot(ee.wavelength, ee.gdd, label=ee.identifier)
    plt.legend()


def plot_layer_thickness_influence(e: Etalon, d_change=1.0):
    n_layer = e.num_layers

    custommap = plt.get_cmap("RdBu")
    plt.figure()
    ax = plt.axes()
    cmm = cm.rainbow
    ax.set_prop_cycle("color", [cmm(i) for i in np.linspace(0, 1, n_layer)])

    norm = matplotlib.colors.BoundaryNorm(np.arange(-0.5, n_layer), custommap.N)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmm)
    e.update()
    wl, design = e.wavelength_vs_residuals_m_per_s()
    for i in range(n_layer):
        corrections = np.zeros_like(e.d_corrections)
        corrections[i] = d_change
        e.d_corrections = corrections
        # e.update()
        plt.plot(wl, design - e.wavelength_vs_residuals_m_per_s()[1])
    plt.title(f"Layer {d_change} nm thickness change - {e.identifier}")
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("Peak shift vs. design values [m/s]")
    cbar = plt.colorbar(mapper, ticks=np.linspace(0, n_layer - 1, n_layer), ax=ax)
    cbar.ax.set_yticklabels([f"{i + 1}" for i in range(n_layer)])
    cbar.set_label("Layer number")


def plot_layer_thickness_influence_gd(e: Etalon, d_change=1.0):
    n_layer = e.num_layers

    custommap = plt.get_cmap("RdBu")
    plt.figure()
    ax = plt.axes()
    cmm = cm.rainbow
    ax.set_prop_cycle("color", [cmm(i) for i in np.linspace(0, 1, n_layer)])

    norm = matplotlib.colors.BoundaryNorm(np.arange(-0.5, n_layer), custommap.N)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmm)
    wl = e.wavelength
    design = e.gd
    for i in range(n_layer):
        pars = np.zeros_like(e.all_parameters)
        pars[i] += d_change
        e.all_parameters = pars

        plt.plot(wl, e.gd - design)
        plt.title(f"Layer {d_change} nm thickness change")
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("GD change")
    cbar = plt.colorbar(mapper, ticks=np.linspace(0, n_layer - 1, n_layer), ax=ax)
    cbar.ax.set_yticklabels([f"{i + 1}" for i in range(n_layer)])
    cbar.set_label("Layer number")

    plt.tight_layout()


def plot_layer_thickness_influence_reflectivity(e: Etalon, d_change=1.0):
    n_layer = e.num_layers

    custommap = plt.get_cmap("RdBu")
    plt.figure()
    ax = plt.axes()
    cmm = cm.rainbow
    ax.set_prop_cycle("color", [cmm(i) for i in np.linspace(0, 1, n_layer)])

    norm = matplotlib.colors.BoundaryNorm(np.arange(-0.5, n_layer), custommap.N)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmm)
    wl = e.wavelength
    design = e.mirror_reflectivity
    plt.plot(wl, design)
    design_params = e.all_parameters.copy()
    for i in range(n_layer):
        pars = design_params.copy()
        pars[i] += d_change
        e.all_parameters = pars

        plt.plot(wl, e.mirror_reflectivity)
        plt.title(f"Layer {d_change} nm thickness change")
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("Reflectivity")
    cbar = plt.colorbar(mapper, ticks=np.linspace(0, n_layer - 1, n_layer), ax=ax)
    cbar.ax.set_yticklabels([f"{i + 1}" for i in range(n_layer)])
    cbar.set_label("Layer number")

    plt.tight_layout()


def plot_reflectivity(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("Mirror Reflectivity")
    for ee in e:
        plt.plot(ee.wavelength, ee.mirror_reflectivity, label=ee.identifier)
    plt.legend()
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("Reflectivity ")


def plot_mirror_transmission(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("Mirror Transmission")
    for ee in e:
        plt.plot(ee.wavelength, ee.mirror_transmission, label=ee.identifier)
    plt.legend()
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("Mirror Transmission ")


def plot_absorption(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("Mirror Absorption")
    for ee in e:
        plt.plot(
            ee.wavelength,
            1.0 - ee.mirror_transmission - ee.mirror_reflectivity,
            label=ee.identifier,
        )
    plt.legend()
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("Mirror Absorption ")


def plot_normalized_fsr(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("Normalized FSR")
    for ee in e:
        # plot the FSR normalized to the nominal FSR
        plt.plot(ee.wavelength, ee.fsr / ee.ideal_fsr, label=ee.identifier)
    plt.legend()
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("Normalized FSR (FSR / ideal FSR)")


def plot_fsr(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("FSR")
    for ee in e:
        plt.plot(ee.wavelength, ee.fsr, label=ee.identifier)
    plt.legend()
    plt.xlabel("Wavelength [nm]")
    plt.ylabel("FSR [GHz]")


def plot_refractive_index_materials(
    etalon, wavelength, exclude_outer_two_materials=True, plot_k=True
):
    wl = wavelength.copy()
    if exclude_outer_two_materials:
        mats = etalon.materials[1:-1].copy()
        mats_names = etalon.names[1:-1].copy()
    else:
        mats = etalon.materials.copy()
        mats_names = etalon.names.copy()

    fig, axes = plt.subplots(max(len(mats), 2), sharex=True)
    fig.suptitle("Refractive indices")
    for i, mat in enumerate(mats):
        (p1,) = axes[i].plot(wl, mat.get_n(wavelength), label="n")
        axes[i].text(0.05, 0.15, mats_names[i], transform=axes[i].transAxes)
        axes[i].set_ylabel("n")

        if plot_k:
            ax_k = axes[i].twinx()
            ax_k.set_ylabel("k")
            (p2,) = ax_k.plot(wl, mat.get_k(wavelength), "g", label="k")
            axes[i].legend([p1, p2], [plot.get_label() for plot in [p1, p2]])

    axes[len(mats) - 1].set_xlabel("Wavelength [nm]")
    # plt.show()


def plot_stack(ee: list[Etalon]):
    minx, maxx = 0, 0
    fig, ax = plt.subplots()
    plt.title("Coating Stack")

    for ii, e in enumerate(ee):
        all_names = np.array([e.names[i] for i in e.idx_stack])

        dlist = e.d_stack_design[1:-1]
        dlist = np.insert(np.append(dlist, 1.3 * np.max(dlist)), 0, np.max(dlist))
        dlist_cumsum = np.cumsum(dlist) - dlist[0]

        left, right = dlist_cumsum[:-1], dlist_cumsum[1:]
        left, right = (
            np.insert(left, 0, -dlist[0]),
            np.append(right, dlist[-1] + dlist_cumsum[-1]),
        )
        bottom, top = np.full(len(left), ii), np.full(len(left), ii + 1)

        colors = ["gray", "lightblue", "lightgreen", "darkgray"]
        unique_names = np.array(e.names)

        for l, r, b, t, name in zip(left, right, bottom, top, all_names):  # noqa: E741
            color_idx = np.where(unique_names == name)[0][0]
            ax.add_patch(
                patches.Rectangle(
                    (l, b), r - l, t - b, edgecolor="k", facecolor=colors[color_idx]
                )
            )
            ax.text(l + 10, b + 0.05, name, rotation=90)

        ax.set_xlabel("Accum. thickness [nm]")
        minx, maxx = min(minx, left[0]), max(maxx, right[-1])

    ax.set_yticks(np.arange(len(ee)) + 0.5)
    ax.set_yticklabels([e.identifier for e in ee])
    ax.set_xlim(minx, maxx)
    ax.set_ylim(0, len(ee))


def plot_reflectivity_finesse(e: Union[Etalon, List[Etalon]]):
    e = [e] if isinstance(e, Etalon) else e
    plt.figure()
    plt.title("Finesse")
    for ee in e:
        plt.plot(ee.wavelength, ee.reflectivity_finesse, label=ee.identifier)
    plt.legend()



from __future__ import annotations

import json
import logging
import pathlib
from functools import cached_property
from typing import Callable

import numpy as np
from pyindexrepo import Material
from scipy.interpolate import UnivariateSpline

from PyEtalon.tmm import rt

# from tmmnlay import calculate_rt as calculate_rt_gpu
logger = logging.getLogger(__name__)
c_handler = logging.StreamHandler()
# logger.addHandler(c_handler)
logger.setLevel("DEBUG")

numba_logger = logging.getLogger("numba")
numba_logger.setLevel(logging.WARNING)

# constants
c = 299792458  # speed of light [m/s]


def guess_m(
    wavelength: float | np.ndarray,
    d: float = 9.99e-3,
    aoi: float = 0.0,
    n: float | Callable = 1.0,
    phase: Callable | None = None,
) -> tuple[int | np.ndarray, np.ndarray]:
    """
    Guesses the peak number m for a given peak wavelength.

    Args:
        wavelength: peak wavelength in [nm]
        d: spacer thickness in [m]
        aoi: angle of incidence in [rad]
        n: refractive index of cavity material
        phase: phase shift upon reflection in [rad]

    Returns:
        guessed peak numbers, difference between integer and float peak numbers
    """
    if phase is None:
        if isinstance(n, float):
            float_m = 2.0 * n * d * np.cos(aoi) / (wavelength * 1e-9)
        elif callable(n):
            float_m = 2.0 * n(wavelength) * d * np.cos(aoi) / (wavelength * 1e-9)
        else:
            raise ValueError("n needs to be a float or a callable function")
        int_m = np.array(np.rint(float_m), dtype=int)
    else:
        if isinstance(n, float):
            float_m = (
                2.0 * d * n * np.cos(aoi)
                + phase(wavelength) * (wavelength * 1e-9) / np.pi
            ) / (wavelength * 1e-9)
        elif callable(n):
            float_m = (
                2.0 * d * n(wavelength.copy()) * np.cos(aoi)
                + phase(wavelength.copy()) * (wavelength * 1e-9) / np.pi
            ) / (wavelength * 1e-9)
        elif isinstance(n, np.ndarray):
            float_m = (
                2.0 * d * n * np.cos(aoi)
                + phase(wavelength.copy()) * (wavelength * 1e-9) / np.pi
            ) / (wavelength * 1e-9)
        else:
            raise ValueError(
                "n needs to be a float, an array of floats or a callable function."
            )
        int_m = np.array(np.rint(float_m), dtype=int)
    return int_m, float_m - int_m


def lambda_peaks(
    m: int | np.ndarray,
    d: float = 9.99e-3,
    n: float | Callable = 1.0,
    aoi: float = 0.0,
    phase: None | Callable = None,
) -> float | np.ndarray:
    """
    Returns peak wavelength for given etalon parameters

    It uses the dispersion formula as given in https://doi.org/10.1364/AO.30.004126
    Args:
        m: Diffraction order (int) or 1D array with 'int' type.
        d: physical mirror distance [m]
        n: Refractive index of medium in between mirrors
        aoi: Angle of incidence [rad]
        phase: Interpolation function f = phase_shift(wavelength[nm])

    Returns:
        peak wavelength in [nm]

    """
    repeat = 3
    # approximate wavelength without phase shift:
    wl = 2.0 * d * 1.0 * np.cos(aoi) / m * 1e9
    if isinstance(n, float):
        wl = 2.0 * d * n * np.cos(aoi) / m * 1e9
    if callable(n):
        for _ in range(repeat):
            wl = 2.0 * d * n(wl) * np.cos(aoi) / m * 1e9
    if phase is not None:
        if isinstance(n, float):
            for _ in range(repeat):
                wl = (2.0 * d * 1e9 * np.cos(aoi) * n + phase(wl) / np.pi * wl) / m
        if callable(n):
            for _ in range(repeat):
                wl = (2.0 * d * 1e9 * np.cos(aoi) * n(wl) + phase(wl) / np.pi * wl) / m
    return wl


class Etalon:
    """
    Class for modeling an etalon with a multilayer thin film stack.

    Attributes:
        materials: List of materials used in the stack
        names: List of names of the materials
        d_stack_design: List of layer thicknesses in [nm]
        d_corrections: List of layer thickness corrections in [nm] - default is an array of zeros (same length as d_stack)
        num_layers: Number of layers in the stack
        identifier: Identifier for the etalon
        wavelength_min: Minimum wavelength in [nm] (if not given, it is calculated from the wavelength vector)
        wavelength_max: Maximum wavelength in [nm] (if not given, it is calculated from the wavelength vector)
        wavelength: Wavelength vector in [nm]
        normalized_wavelength: Normalized wavelength vector
        aoi: Angle of incidence in [rad]
        _d_spacer: Spacer thickness in [m] (incl. corrections if given)
        d_spacer_correction: Spacer thickness correction in [nm] - default is 0.0
        idx_stack: List of indices of the materials in the stack
        _m: Peak order number
    """

    def __init__(
        self,
        materials: list[Material],
        material_names: list[str] | None,
        d_stack: list[float] | np.ndarray,
        idx_stack: list[int],
        wavelength_min: float | None = None,
        wavelength_max: float | None = None,
        wavelength: list[float] | np.ndarray = None,
        d_spacer: float = 10e-3,
        aoi: float = 0.0,
        identifier: str = "Etalon",
    ):
        """Constructor for the Etalon class

        Args:
            materials: List of materials used in the stack
            material_names: List of names of the materials
            d_stack: List of layer thicknesses in [nm]
            idx_stack: List of indices of the materials in the stack
            wavelength_min: Minimum wavelength in [nm] (if not given, it is calculated from the wavelength vector)
            wavelength_max: Maximum wavelength in [nm] (if not given, it is calculated from the wavelength vector)
            wavelength: Wavelength vector in [nm]
            d_spacer: Spacer thickness in [m] (incl. corrections if given)
            aoi: Angle of incidence in [deg]
            identifier: Identifier for the etalon
        """
        if material_names is None:
            material_names = [m.name for m in materials]
        assert len(materials) == len(material_names), (
            "Materials and Material_names need to be the same length."
        )
        assert len(d_stack) == len(idx_stack), (
            f"Length of d_stack ({len(d_stack)}) and idx_stack ({len(idx_stack)}) is not equal."
        )

        assert wavelength is not None or (
            wavelength_min is not None and wavelength_max is not None
        ), (
            "Either specify the wavelength vector for calculations explicitly or specify minimum and maximum wavelength"
        )

        self.materials = materials
        self.names = material_names
        self.d_stack_design = np.array(d_stack)
        self.d_corrections = np.zeros_like(
            d_stack[1:-1]
        )  # exclude the spacer and ambient material
        self.num_layers = len(d_stack) - 2  # exclude the spacer and ambient material
        self.identifier = identifier
        if wavelength is not None:
            self.wavelength = wavelength
            self.wavelength_min = np.min(wavelength)
            self.wavelength_max = np.max(wavelength)
            self.normalized_wavelength = np.linspace(0, 1, len(wavelength))

        else:
            self.wavelength_min = wavelength_min
            self.wavelength_max = wavelength_max
            self.wavelength = np.linspace(wavelength_min, wavelength_max, 5000)
            self.normalized_wavelength = np.linspace(0, 1, len(self.wavelength))

        self.aoi = aoi
        self._d_spacer = d_spacer
        self.d_spacer_correction = 0.0

        self.idx_stack = np.array(idx_stack, dtype=int)

        self.material_coefficients = []
        for i, m in enumerate(materials[1:-1]):
            # check if material has attribute 'coefficients' - could be TabulatedIndexData
            if hasattr(m.n, "coefficients"):
                self.material_coefficients.append(m.n.coefficients)
            if m.k is not None:
                if hasattr(m.k, "coefficients"):
                    self.material_coefficients.append(m.k.coefficients)

        # Convert given wavelength to peak order number
        # self._m = guess_m(self.wavelength, self.d_spacer, self.aoi, self.n(0), self.phase_spline)
        self._m = np.arange(
            int(
                (
                    2.0
                    * self.d_spacer
                    * self.get_refractive_index(0, self.wavelength_max)
                )
                / (self.wavelength_max * 1e-9)
            ),
            int(
                (
                    2.0
                    * self.d_spacer
                    * self.get_refractive_index(0, self.wavelength_min)
                )
                / (self.wavelength_min * 1e-9)
            ),
        )[::-1]

    def __str__(self):
        return f"{self.identifier} Etalon: \n Number of layers: {self.num_layers - 2}. \n Used materials are: {self.materials}"

    def json_dict(self):
        json_dict = {}
        for k in [
            # "materials",
            # "names",
            # "d_stack",
            # "idx_stack",
            # "wavelength_min",
            # "wavelength_max",
            # "d_spacer",
            # "aoi",
            # "identifier",
            "all_parameters",
        ]:
            val = getattr(self, k)
            # TODO: make it work for materials
            if isinstance(val, np.ndarray):
                json_dict[k] = val.tolist()
            else:
                json_dict[k] = getattr(self, k)
        return json_dict

    def save_to_json(self, filename):
        with open(filename, "w") as f:
            json.dump(self.json_dict(), f, indent=4)

    def get_refractive_index(self, idx_mat, wl=None):
        if wl is None:
            return self.materials[idx_mat].get_n(
                np.atleast_1d(self.wavelength / 1000.0)
            )
        else:
            return self.materials[idx_mat].get_n(np.atleast_1d(wl / 1000.0))

    def get_refractive_index_cavity(self, wl):
        """Returns the refractive index of the cavity material"""
        return self.get_refractive_index(0, wl)

    @property
    def rt(self):
        """Returns the reflection and transmission coefficients of the mirror coating stack"""
        return rt(self.nk_stack, self.d_stack, self.wavelength, self.aoi)

    @property
    def d_spacer(self) -> float:
        """Distance of spacer in [m] (incl. corrections if given)"""
        return self._d_spacer + self.d_spacer_correction / 1e9

    @property
    def nk_stack(self):
        # Calculate refractive indices
        res = np.empty(
            (len(self.wavelength), len(self.d_stack_design)), dtype=np.complex128
        )

        for i in np.sort(np.unique(self.idx_stack)):
            nn = self.get_refractive_index(i, self.wavelength)
            kk = self.materials[i].get_k(self.wavelength / 1000.0)
            nk_data = np.vectorize(complex)(nn, kk)
            for j, idx_mat in enumerate(self.idx_stack):
                if idx_mat == i:
                    res[:, j] = nk_data

        return res

    @cached_property
    def num_material_coefficients(self):
        """Returns the total number of coefficients for all coating materials

        The substrate and ambient material are not included.
        """
        n = 0
        for m in self.materials[1:-1]:
            if hasattr(m.n, "coefficients"):
                n += m.n.coefficients.size
            if m.k is not None:
                if hasattr(m.k, "coefficients"):
                    n += m.k.coefficients.size
        return n

    @cached_property
    def n_parameter(self):
        """Total number of parameters

        Number of layers + 1 for the spacer + number of coefficients for each coating material

        """
        return self.num_layers + 1 + self.num_material_coefficients

    @cached_property
    def all_parameter_names(self):
        names = ["d_spacer"]
        names.extend([f"d_{i}" for i in range(self.d_stack.size - 2)])
        for i, m in enumerate(self.materials[1:-1]):
            for j in range(m.n.coefficients.size):
                names.append(f"{self.materials[i + 1].name}_n_{j}")
            if m.k is not None:
                for j in range(m.k.coefficients.size):
                    names.append(f"{self.materials[i + 1].name}_k_{j}")
        return names

    @property
    def all_parameters(self):
        pars = np.zeros(self.n_parameter)
        pars[0] = self.d_spacer_correction
        pars[1 : self.num_layers + 1] = self.d_corrections
        j = 0
        for i, m in enumerate(self.materials[1:-1]):
            if hasattr(m.n, "coefficients"):
                n = m.n.coefficients.size
                pars[self.num_layers + 1 + j : self.num_layers + 1 + j + n] = (
                    m.n.coefficients
                )
                j += n
            if m.k is not None:
                if hasattr(m.k, "coefficients"):
                    n = m.k.coefficients.size
                    pars[self.num_layers + 1 + j : self.num_layers + 1 + j + n] = (
                        m.k.coefficients
                    )
                    j += n
        return pars

    def print_parameters(self):
        """print parameters names and parameters and index in a table"""
        print(f"{'Index':<5} {'Parameter':<40} {'Value':<20}")
        print("-" * 65)
        for i, (name, par) in enumerate(
            zip(self.all_parameter_names, self.all_parameters)
        ):
            print(f"{i:<5} {name:<40} {par:<20}")

    def print_parameters_horizontally(self, values_only=False):
        """print index, parameter names and values in a table.
        First row contains index, second row contains name, third row contains all values.
        Fix the column width to 20 characters"""
        if not values_only:
            print(f"{'Index':<10}", end="")
            for i in range(self.n_parameter):
                print(f"{i:<6}", end="")
            print()
            print("-" * 65)
            print(f"{'Parameter':<10}", end="")
            for name in self.all_parameter_names:
                print(f"{name:<7}", end="")
            print()
            print("-" * 65)
        print(f"{'Value':<10}", end="")
        for par in self.all_parameters:
            print(f"{par:7.3f}", end="")
        print()

    @all_parameters.setter
    def all_parameters(self, pars):
        assert pars.size == self.n_parameter, (
            f"Number of parameters does not match. Expected {self.n_parameter}, got {pars.size}"
        )

        self.d_spacer_correction = pars[0]
        self.d_corrections = pars[1 : self.num_layers + 1]

        j = 0
        for i, m in enumerate(self.materials[1:-1]):
            if hasattr(m.n, "coefficients"):
                n = m.n.coefficients.size

                m.n.coefficients = pars[
                    self.num_layers + 1 + j : self.num_layers + 1 + j + n
                ]
                j += n
            if m.k is not None:
                if hasattr(m.k, "coefficients"):
                    n = m.k.coefficients.size

                    m.k.coefficients = pars[
                        self.num_layers + 1 + j : self.num_layers + 1 + j + n
                    ]
                    j += n
        # self.update()

    @property
    def d_stack(self):
        """Returns the stack thickness including any corrections given (if any)"""
        dstack = self.d_stack_design.copy()
        dstack[1:-1] += self.d_corrections
        return dstack

    @property
    def d_stack_normalized(self):
        """Returns the stack thickness normalized to the maximum thickness"""
        stack_max_thickness = np.max(self.d_stack[1:-1])
        return self.d_stack / stack_max_thickness

    @property
    def phase(self):
        """Returns the unwrapped phase shift upon reflection in [rad] as a function of wavelength"""
        return np.unwrap(np.angle(self.rt[0]))

    @property
    def phase_spline(self):
        """Returns the phase as a spline"""
        return UnivariateSpline(self.wavelength, self.phase, k=3, s=0)

    @property
    def gd(self):
        """Returns the group delay upon reflection in [fs]"""
        return (
            self.phase_spline.derivative(1)(self.wavelength)
            * 1e6
            * self.wavelength**2
            / (2.0 * np.pi * c)
        )

    @property
    def gd_spline(self):
        """Returns the group delay upon reflection in [fs] as a spline"""
        return UnivariateSpline(self.wavelength, self.gd, k=5, s=0.0)

    @property
    def gdd(self):
        """Returns the group delay dispersion upon reflection in [fs^2]"""
        return (
            self.phase_spline.derivative(2)(self.wavelength)
            * self.wavelength**3
            / (2.0 * np.pi)
            / 1000.0
        )

    @property
    def gdd_spline(self):
        """Returns the group delay dispersion upon reflection in [fs^2] as a spline"""
        return UnivariateSpline(self.wavelength, self.gdd, k=5, s=0)

    def calculate_reflectivity_transmissivity(self) -> tuple[np.ndarray, np.ndarray]:
        """Calculate reflectivity and transmissivity from reflection and transmission coefficients.

        Returns:
            tuple[np.ndarray, np.ndarray]:ay of float64 values.
                - Transmissivity (T): 1D array of float64 values.
        """
        # Convert angle of incidence to radians
        aoi_rad = np.radians(self.aoi)

        # Extract refractive indices at the boundaries
        n_i = self.nk_stack[:, 0]  # Incident medium
        n_t = self.nk_stack[:, -1]  # Transmitted medium

        # Calculate transmitted angle using Snell's law: n_i * sin(theta_i) = n_t * sin(theta_t)
        sin_theta_t = np.real(n_i) * np.sin(aoi_rad) / np.real(n_t)
        cos_theta_t = np.sqrt(
            1 - sin_theta_t**2
        )  # cos(theta_t), assuming real refractive indices
        cos_theta_i = np.cos(aoi_rad)  # cos(theta_i)
        r, t = self.rt
        # Reflectivity: R = |r|^2
        R = np.abs(r) ** 2

        # Transmissivity: T = |t|^2 * Re(n_t cos(theta_t)) / Re(n_i cos(theta_i))
        T = (
            (np.abs(t) ** 2)
            * (np.real(n_t) * cos_theta_t)
            / (np.real(n_i) * cos_theta_i)
        )

        return R, T

    @property
    def mirror_reflectivity(self):
        """Returns the reflectivity of the mirror coating stack"""
        return self.calculate_reflectivity_transmissivity()[0]

    @property
    def mirror_transmission(self):
        """Returns the transmission of the mirror coating stack"""
        return self.calculate_reflectivity_transmissivity()[1]

    @property
    def fsr(self):
        """Return the real free spectral range in [GHz]

        The free spectral range is given by the formula:

        FSR [GHz] = c / (2 * n * d * cos(theta) - lambda^2 * 10^-9 / pi * dphi/dlambda)

        see https://doi.org/10.1103/PhysRevA.37.1802

        where:
        c: speed of light [m/s]
        n: refractive index of cavity material
        d: spacer thickness [m]
        theta: angle of incidence [rad]
        lambda: peak wavelength [nm]
        dphi/dlambda: derivative of phase with respect to wavelength [rad/nm]

        The factor 10^-9 is used to convert the wavelength from nm to m. There is a factor of 10^-18 from lambda^2, but
        also a factor of 10^9 from the derivative of the phase. The factor of 10^-9 cancels out.

        """
        n = self.get_refractive_index_cavity(self.wavelength)
        return (
            c
            / (
                2.0 * n * self.d_spacer * np.cos(np.deg2rad(self.aoi))
                - (self.wavelength**2 / np.pi)
                * 1e-9
                * self.phase_spline.derivative(1)(self.wavelength)
            )
            / 1e9
        )

    @property
    def ideal_fsr(self) -> float:
        """Return the ideal free spectral range in [GHz]

        The ideal free spectral range is given by the formula:

        FSR [GHz] = c / (2 * n * d * cos(theta))

            where:
            c: speed of light [m/s]
            n: refractive index of cavity material
            d: spacer thickness [m]
            theta: angle of incidence [rad]

        It neglects the phase shift upon reflection.
        """
        return (
            c
            / (
                2.0
                * self.get_refractive_index_cavity(self.wavelength).mean()
                * self.d_spacer
                * np.cos(np.deg2rad(self.aoi))
            )
            / 1e9
        )

    @property
    def fsr_wavelength(self):
        """Return the free spectral range in [nm]

        ChatGPT gives me this as approximation (phase shift is a small correction to optical path)

        FSR_lambda = lambda^2 / (2nL cos(theta)) (1+ (lambda^2/(2pi n L cos(theta)) dPhi/dlambda))
        """
        n = self.get_refractive_index_cavity(self.wavelength)
        # First, compute the base term: lambda^2 in meters
        lambda_sq_m2 = (self.wavelength * 1e-9) ** 2
        cos_theta = np.cos(np.deg2rad(self.aoi))

        # Base FSR in meters
        base_fsr_m = lambda_sq_m2 / (2 * n * self.d_spacer * cos_theta)

        # Correction term due to mirror phase dispersion
        correction_factor = (
            lambda_sq_m2 / (2 * np.pi * n * self.d_spacer * cos_theta)
        ) * self.phase_spline.derivative(1)(self.wavelength)

        # FSR in meters including phase correction
        fsr_m = base_fsr_m * (1 + correction_factor)

        # Convert result to nanometers
        return fsr_m * 1e9

    @property
    def coefficient_of_finesse(self):
        """Returns the coefficient of finesse of the mirror coating stack

        The coefficient of finesse is given by the formula:

        F = 4 * R / (1 - R)^2

        where:
        R: mirror reflectivity

        """
        return 4.0 * self.mirror_reflectivity / (1.0 - self.mirror_reflectivity) ** 2

    @property
    def reflectivity_finesse(self):
        """Returns the reflectivity finesse of the mirror coating stack

        The reflectivity finesse is given by the formula:

        F = pi / (2 * arcsin(1 / sqrt(F)))

        where:
        F: coefficient of finesse

        """
        return np.pi / (2.0 * np.arcsin(1.0 / np.sqrt(self.coefficient_of_finesse)))

    def peak_wavelength_ideal(self):
        return lambda_peaks(
            self._m,
            d=self.d_spacer,
            n=self.get_refractive_index_cavity,
            aoi=np.deg2rad(self.aoi),
        )

    def peak_wavelength_real(self, m=None):
        if m is None:
            m = self._m
        return lambda_peaks(
            m,
            d=self.d_spacer,
            n=self.get_refractive_index_cavity,
            aoi=np.deg2rad(self.aoi),
            phase=self.phase_spline,
        )

    def residuals_ideal_vs_real(self):
        return self.peak_wavelength_ideal() - self.peak_wavelength_real()

    def set_data(self, wavelength):
        self.wavelength = wavelength
        self.normalized_wavelength = np.linspace(0, 1, len(wavelength))
        self._m = guess_m(
            wavelength,
            self.d_spacer,
            self.aoi,
            n=self.get_refractive_index_cavity,
            phase=self.phase_spline,
        )[0]

    def update(self):
        self._m = guess_m(
            self.wavelength,
            self.d_spacer,
            self.aoi,
            n=self.get_refractive_index_cavity,
            phase=self.phase_spline,
        )[0]

    def measured_residuals(self):
        return (
            (self.wavelength - self.peak_wavelength_real())
            / self.peak_wavelength_real()
            * c
        )

    def wavelength_vs_residuals_m_per_s(self):
        return (
            self.peak_wavelength_ideal(),
            self.residuals_ideal_vs_real() / self.peak_wavelength_ideal() * c,
        )

    def inc_m(self):
        self._m += 1

    def dec_m(self):
        self._m -= 1

    def guess_m(self, measured_wl, phase=None):
        if phase is None:
            phase = self.phase_spline
        return guess_m(
            measured_wl,
            self.d_spacer,
            np.deg2rad(self.aoi),
            self.get_refractive_index_cavity,
            phase=phase,
        )

    def transmission_spectrum(self, wavelength):
        """Returns the transmission spectrum of the entire etalon for the given wavelength.

        The transmission spectrum is calculated as:

        T = 1 / (1 + F * sin^2(pi * d / lambda + phi))

        where:
        T: transmission spectrum
        F: coefficient of finesse
        d: spacer thickness
        lambda: wavelength
        phi: phase shift upon reflection

        Args:
            wavelength: Wavelength in [nm]

        """
        F_spline = UnivariateSpline(
            self.wavelength, self.coefficient_of_finesse, k=3, s=0
        )

        return 1.0 / (
            1.0
            + F_spline(wavelength)
            * np.sin(np.pi * self.d_spacer * 1e9 / wavelength) ** 2
        )

    def load_parameters(self, filename: str | pathlib.Path):
        with open(filename, "r") as f:
            pars = json.load(f)
        self.all_parameters = np.array(pars["all_parameters"])


# def load_from_file(filename):
#     with open(filename, "r") as json_file:
#         return Etalon(**json.load(json_file))



import numpy as np
from numba import njit


@njit(cache=True)
def rt(
    n: np.ndarray, d: np.ndarray, wvl: np.ndarray, aoi: float = 0.0, pol: int = 0
) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculate the reflection and transmission coefficients for a multilayer.

    Parameters:
        n: 2D array-like, shape (num_wavelengths, num_layers)
            Refractive indices for each layer and wavelength.
        d: 1D array-like, shape (num_layers)
            Thickness of each layer (units must match wavelength).
        wvl: 1D array-like, shape (num_wavelengths)
            Wavelengths of light (units must match thickness).
        aoi: float, optional
            Angle of incidence in degrees. Default is 0.0.
        pol: int, optional
            Polarization state, 0 for TE (s) or 1 for TM (p). Default is 0.

    Returns:
        r: 1D array, shape (num_wavelengths)
            Reflection coefficients for each wavelength.
        t: 1D array, shape (num_wavelengths)
            Transmission coefficients for each wavelength.
    """

    # Calculate squared sine of angle of incidence
    sin_squared_aoi = np.sin(np.radians(aoi)) ** 2
    n_squared = n**2

    # Precompute interface matrices
    n_current_layer, n_next_layer = n[:, :-1], n[:, 1:]
    s2_per_layer = (n_squared[:, 0:1] * sin_squared_aoi) / n_squared
    cos_current_layer, cos_next_layer = (
        np.sqrt(1 - s2_per_layer[:, :-1]),
        np.sqrt(1 - s2_per_layer[:, 1:]),
    )

    num1 = n_current_layer * cos_current_layer + n_next_layer * cos_next_layer
    num2 = n_current_layer * cos_next_layer + n_next_layer * cos_current_layer

    rjk_TE = (
        n_current_layer * cos_current_layer - n_next_layer * cos_next_layer
    ) / num1
    tjk_TE = 2.0 * n_current_layer * cos_current_layer / num1
    rjk_TM = (
        n_current_layer * cos_next_layer - n_next_layer * cos_current_layer
    ) / num2
    tjk_TM = 2.0 * n_current_layer * cos_next_layer / num2

    rjk = rjk_TE if pol == 0 else rjk_TM
    tjk = tjk_TE if pol == 0 else tjk_TM

    # Precompute layer matrices
    wave_vector_z = (
        np.sqrt(n_squared[:, 1:-1] - n_squared[:, 0:1] * sin_squared_aoi) * d[1:-1]
    )
    Bj = 2j * np.pi * wave_vector_z / wvl[:, None]
    exp_neg_Bj, exp_pos_Bj = np.exp(-Bj), np.exp(Bj)

    # Total transfer matrix initialization
    S11, S12 = 1.0 / tjk[:, 0], rjk[:, 0] / tjk[:, 0]
    S21, S22 = S12, S11

    for j in range(d.shape[0] - 2):  # Loop over middle layers
        B11, B12 = (
            exp_neg_Bj[:, j] / tjk[:, j + 1],
            exp_neg_Bj[:, j] * rjk[:, j + 1] / tjk[:, j + 1],
        )
        B21, B22 = (
            exp_pos_Bj[:, j] * rjk[:, j + 1] / tjk[:, j + 1],
            exp_pos_Bj[:, j] / tjk[:, j + 1],
        )

        # Update total matrix S = S * B
        C11 = S11 * B11 + S12 * B21
        C12 = S11 * B12 + S12 * B22
        C21 = S21 * B11 + S22 * B21
        C22 = S21 * B12 + S22 * B22

        S11, S12, S21, S22 = C11, C12, C21, C22

    # Final calculation for reflection and transmission
    r = S21 / S11
    t = 1.0 / S11

    return r, t


#
#
# @cuda.jit
# def rt_cuda(n, d, wvl, aoi, pol, r_real, r_imag, t_real, t_imag):
#     """
#     CUDA kernel to calculate the reflection and transmission coefficients for a multilayer.
#
#     Parameters:
#         n: 2D device array, shape (num_wavelengths, num_layers)
#             Refractive indices for each layer and wavelength.
#         d: 1D device array, shape (num_layers)
#             Thickness of each layer (units must match wavelength).
#         wvl: 1D device array, shape (num_wavelengths)
#             Wavelengths of light (units must match thickness).
#         aoi: float
#             Angle of incidence in degrees.
#         pol: int
#             Polarization state, 0 for TE (s) or 1 for TM (p).
#         r_real, r_imag: 1D device arrays
#             Output arrays for real and imaginary parts of reflection coefficients.
#         t_real, t_imag: 1D device arrays
#             Output arrays for real and imaginary parts of transmission coefficients.
#     """
#     idx = cuda.grid(1)
#     if idx < wvl.shape[0]:
#         # Constants
#         s2 = (np.sin(aoi * np.pi / 180))**2  # Squared sine of the angle of incidence
#         num_layers = d.shape[0]
#
#         # Precompute n^2 and angle terms
#         n2 = cuda.local.array(32, dtype=np.complex128)  # Max layers = 32
#         for j in range(num_layers):
#             n2[j] = n[idx, j] * n[idx, j]
#         s2_per_layer = (n2[0] * s2) / n2
#
#         # Compute cosines for layers
#         cj = cuda.local.array(32, dtype=np.float32)
#         for j in range(num_layers):
#             cj[j] = np.sqrt(1.0 - s2_per_layer[j])
#
#         # Initialize S-matrix
#         S11_real, S11_imag = 1.0, 0.0
#         S21_real, S21_imag = 0.0, 0.0
#
#         for j in range(num_layers - 1):
#             nj, nk = n[idx, j], n[idx, j + 1]
#             cj_j, cj_k = cj[j], cj[j + 1]
#
#             # Fresnel coefficients
#             num1 = nj * cj_j + nk * cj_k
#             num2 = nj * cj_k + nk * cj_j
#
#             if pol == 0:  # TE polarization
#                 rjk = (nj * cj_j - nk * cj_k) / num1
#                 tjk = 2.0 * nj * cj_j / num1
#             else:  # TM polarization
#                 rjk = (nj * cj_k - nk * cj_j) / num2
#                 tjk = 2.0 * nj * cj_k / num2
#
#             # Phase term
#             if j > 0:
#                 kj = np.sqrt(n2[j] - n2[0] * s2) * d[j]
#                 phase = 2.0 * np.pi * kj / wvl[idx]
#                 cos_phase = np.cos(phase)
#                 sin_phase = np.sin(phase)
#             else:
#                 cos_phase = 1.0
#                 sin_phase = 0.0
#
#             # Update S-matrix
#             T11_real = cos_phase / tjk
#             T11_imag = -sin_phase / tjk
#             T21_real = T11_real * rjk
#             T21_imag = T11_imag * rjk
#
#             S11_new_real = S11_real * T11_real - S11_imag * T11_imag
#             S11_new_imag = S11_real * T11_imag + S11_imag * T11_real
#             S21_new_real = S21_real * T11_real - S21_imag * T11_imag
#             S21_new_imag = S21_real * T11_imag + S21_imag * T11_real
#
#             S11_real, S11_imag = S11_new_real, S11_new_imag
#             S21_real, S21_imag = S21_new_real, S21_new_imag
#
#         # Final reflection and transmission coefficients
#         r_real[idx] = S21_real / (S11_real**2 + S11_imag**2)
#         r_imag[idx] = S21_imag / (S11_real**2 + S11_imag**2)
#         t_real[idx] = S11_real / (S11_real**2 + S11_imag**2)
#         t_imag[idx] = S11_imag / (S11_real**2 + S11_imag**2)
#
# def rt_host(n, d, wvl, aoi=0.0, pol=0):
#     num_wavelengths = wvl.shape[0]
#
#     # Allocate device memory
#     n_device = cuda.to_device(n.astype(np.complex128))
#     d_device = cuda.to_device(d.astype(np.float32))
#     wvl_device = cuda.to_device(wvl.astype(np.float32))
#     r_real_device = cuda.device_array(num_wavelengths, dtype=np.float32)
#     r_imag_device = cuda.device_array(num_wavelengths, dtype=np.float32)
#     t_real_device = cuda.device_array(num_wavelengths, dtype=np.float32)
#     t_imag_device = cuda.device_array(num_wavelengths, dtype=np.float32)
#
#     # Configure CUDA kernel
#     threads_per_block = 256
#     blocks_per_grid = (num_wavelengths + threads_per_block - 1) // threads_per_block
#
#     # Launch kernel
#     rt_cuda[blocks_per_grid, threads_per_block](
#         n_device, d_device, wvl_device, aoi, pol,
#         r_real_device, r_imag_device, t_real_device, t_imag_device
#     )
#
#     # Copy results back to host
#     r_real = r_real_device.copy_to_host()
#     r_imag = r_imag_device.copy_to_host()
#     t_real = t_real_device.copy_to_host()
#     t_imag = t_imag_device.copy_to_host()
#
#     # Combine real and imaginary parts into complex arrays
#     r = r_real + 1j * r_imag
#     t = t_real + 1j * t_imag
#
#     return r, t
#



import numpy as np

from PyEtalon.tmm import rt


def test_tmm():
    n = np.array([[1.0, 1.5]], dtype=np.complex128)
    d = np.array([-np.inf, np.inf])
    wl = np.array([0.5])
    aoi = 20.0
    r, t = rt(n, d, wl, aoi, 0)
    R = np.abs(r) ** 2

    assert np.isclose(R, 0.0471, atol=1e-4)

    # test other polarization
    r, t = rt(n, d, wl, aoi, 1)
    R = np.abs(r) ** 2
    assert np.isclose(R, 0.0335, atol=1e-4)


if __name__ == "__main__":
    test_tmm()
